use std::mem;
use std::pin::Pin;
use std::task::{Context, Poll, Waker};
use std::sync::{Arc, Mutex, MutexGuard};

use futures::prelude::*;
use futures::ready;
use futures_signals::signal::{Mutable, Signal, MutableSignalCloned, MutableLockMut};

use termion::event::Key;

use crate::input::Inputs;
use crate::schema::API;

use slog::{
    Logger,
    Drain,
    Level,
    Record,
    OwnedKVList,
};

use tui::text::{Span, Spans};
use tui::style::{Style, Color};

#[derive(PartialEq, Eq, Debug, Clone)]
pub enum Window {
    Main,
    Help,
}

/// Application state struct
pub struct Sute<'a, S> {

    // TODO: BE SMART. Inputs change the state, resize signals change the state, futures completing
    // change the state.

    state: SuteState<'a>,

    signal: S,
    inputs: Inputs,
    api: Option<API>,
    log: Logger,
}

impl<'a, S: Unpin> Sute<'a, S> {
    pub fn new(s: S, log: Logger, drain: Arc<LogDrain<'a>>, api: API) -> Self {
        let inputs = Inputs::new();
        let state = SuteState::new(drain);

        Self { 
            state: state,

            signal: s,
            inputs: inputs,
            api: Some(api),
            log: log,
        }
    }

    fn run_cmd(&mut self, cmdline: String) {
        let mut words = cmdline.split_ascii_whitespace();
        match words.next() {
            Some("quit") => self.state.running = false,
            Some("connect") => {
                self.connect(cmdline)
            },
            cmd => info!(self.log, "Issues unknown cmd: {:?}", cmd),
        }
    }

    fn connect(&mut self, params: String) {
        info!(self.log, "Called connect with {}", params);
    }

    fn handle_resize(&mut self, new_size: (u16,u16)) {
        self.state.size = new_size;
    }

    pub fn get_state(&self) -> SuteState<'a> {
        self.state.clone()
    }

    pub fn handle_key(&mut self, key: Key) {
        match key {
            Key::Char('\n') => {
                let cmd = mem::replace(&mut self.state.cmd_line, String::new());
                self.run_cmd(cmd);
            },
            Key::Char(c) => {
                self.state.cmd_line.push(c);
            },
            Key::Backspace => {
                self.state.cmd_line.pop();
            },
            _ => {}
        }
    }
}

impl<'a, S: Signal<Item=(u16,u16)> + Unpin> Signal for Sute<'a, S> {
    type Item = SuteState<'a>;
    fn poll_change(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Option<Self::Item>> {
        let mut dirty = false;

        if let Poll::Ready(Some(size)) = Pin::new(&mut self.signal).poll_change(cx) {
            dirty = true;
            self.handle_resize(size);
        }
        match ready!(Pin::new(&mut self.inputs).poll_next(cx)) {
            Some(key) => {
                dirty = true;
                self.handle_key(key);
            },
            // If the input closes stop the program
            None => {
                self.state.running = false;
                return Poll::Ready(None);
            },
        }

        if dirty {
            Poll::Ready(Some(self.get_state()))
        } else {
            Poll::Pending
        }
    }
}

impl <'a, S> Future for Sute<'a, S> {
    type Output = ();
    fn poll(self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
        if self.state.running {
            Poll::Pending
        } else {
            Poll::Ready(())
        }
    }
}

#[derive(Debug, Clone)]
// TODO: `Signal` struct changes using this struct?
// TODO: If so, procmacro here?
enum SuteStateChange {
    active_win(Window),
}

#[derive(Debug, Clone)]
pub struct SuteState<'a> {
    pub active_win: Window,
    pub size: (u16,u16),
    pub running: bool,
    pub tick: usize,
    pub server: Option<&'a str>,
    pub log: Arc<LogDrain<'a>>,
    pub cmd_line: String,
    pub tick_c: char,
}

impl<'a> SuteState<'a> {
    pub fn new(log: Arc<LogDrain<'a>>) -> Self {
        Self {
            active_win: Window::Main,
            size: (80,20),
            running: true,
            tick: 0,
            server: None,
            log: log,
            cmd_line: String::new(),
            tick_c: '|',
        }
    }
}

#[derive(Debug)]
pub struct LogDrain<'a> {
    inner: Mutex<Vec<Spans<'a>>>,
}

impl<'a> LogDrain<'a> {
    pub fn new() -> Self {
        Self { inner: Mutex::new(Vec::new()) }
    }
    pub fn get_inner(&self) -> MutexGuard<Vec<Spans<'a>>> {
        self.inner.lock().unwrap()
    }
}
impl<'a> Drain for LogDrain<'a> {
    type Ok = ();
    type Err = ();
    fn log(&self, record: &Record, values: &OwnedKVList) -> Result<Self::Ok, Self::Err> {
        let critical_style: Style = Style::default().fg(Color::Magenta);
        let error_style: Style = Style::default().fg(Color::Red);
        let warning_style: Style = Style::default().fg(Color::Yellow);
        let info_style: Style = Style::default().fg(Color::Cyan);
        let debug_style: Style = Style::default().fg(Color::White);
        let trace_style: Style = Style::default().fg(Color::Green);

        let lvlspan = match record.level() {
            Level::Critical => Span::styled("CRIT", critical_style),
            Level::Error => Span::styled("ERRO", error_style),
            Level::Warning => Span::styled("WARN", warning_style),
            Level::Info => Span::styled("INFO", info_style),
            Level::Debug => Span::styled("DEBG", debug_style),
            Level::Trace => Span::styled("TRCE", trace_style),
        };
        let contentspan = Span::from(record.msg().to_string());

        {
            let mut v = self.inner.lock().unwrap();
            v.push(Spans::from(vec![
                Span::from("["), lvlspan, Span::from("]: "),
                contentspan,
            ]));
        }

        Ok(())
    }
}

